#!/usr/bin/env python
# coding: utf-8

# In[ ]:


# treemun_sim/core.py
"""
Función principal del simulador forestal.
"""

import pandas as pd
import numpy as np
import random
from typing import List, Tuple, Dict, Any
import pkg_resources
from .generadores import generar_rodales_aleatorios, generar_rodales, generar_rodalesconpolicy
from .simulacion import simula_bosque as simula_bosque_interno, getBiomasa4Opti

# Semilla global para reproducibilidad
SEMILLA_GLOBAL = 5555

def cargar_lookup_table():
    """Carga la tabla de lookup incluida en el paquete."""
    try:
        ruta_archivo = pkg_resources.resource_filename('treemun_sim', 'data/lookup_table.csv')
        df = pd.read_csv(ruta_archivo, keep_default_na=False)
        df.set_index(
            ["Especie", "Zona", "SiteIndex", "Manejo", "Condicion", "DensidadInicial"],
            inplace=True
        )
        dict_idx = df["id"].to_dict()
        return df, dict_idx
    except Exception as e:
        raise FileNotFoundError(f"No se pudo cargar lookup_table.csv: {e}")

def simular_bosque(
    policies_pino: List[Tuple[int, int]] = None,
    policies_eucalyptus: List[Tuple[int,]] = None,
    horizonte: int = 30,
    num_rodales: int = 100,
    semilla: int = None
) -> Tuple[List, List, Dict, Dict]:
    """
    Función principal del simulador forestal.
    
    Args:
        policies_pino: Lista de políticas para pino, formato [(raleo, cosecha), ...]
                      Por defecto: [(9, 18), (9, 20), ..., (12, 24)]
        policies_eucalyptus: Lista de políticas para eucalipto, formato [(cosecha,), ...]
                            Por defecto: [(9,), (10,), (11,), (12,)]
        horizonte: Horizonte temporal de la simulación (años)
        num_rodales: Número de rodales a generar
        semilla: Semilla para reproducibilidad. Si es None, usa SEMILLA_GLOBAL (5555)
        
    Returns:
        Tupla con (bosque, resumen, biomasa_final_por_rodal, biomasa_estimada)
    """
    
    # Establecer semilla (usar valor por defecto si no se especifica)
    if semilla is None:
        semilla = SEMILLA_GLOBAL
    
    np.random.seed(semilla)
    random.seed(semilla)
    
    # Políticas por defecto si no se especifican
    if policies_pino is None:
        policies_pino = [
            (9, 18), (9, 20), (9, 22), (9, 24),
            (10, 18), (10, 20), (10, 22), (10, 24),
            (11, 18), (11, 20), (11, 22), (11, 24),
            (12, 18), (12, 20), (12, 22), (12, 24)
        ]
    
    if policies_eucalyptus is None:
        policies_eucalyptus = [(9,), (10,), (11,), (12,)]
    
    # Cargar datos base
    df, dict_idx = cargar_lookup_table()
    
    # Generar configuración de rodales aleatorios
    config = generar_rodales_aleatorios(
        df=df,
        dict_idx=dict_idx,
        num_rodales=num_rodales,
        horizonte=horizonte,
        policies_pino=policies_pino,
        policies_eucalyptus=policies_eucalyptus
    )
    
    # Generar rodales base
    rodales = generar_rodales(config, df, dict_idx)
    
    # Aplicar políticas
    rodales_con_policy = generar_rodalesconpolicy(rodales, config)
    
    # Simular crecimiento
    bosque, resumen, biomasa_final_por_rodal = simula_bosque_interno(
        rodales_con_policy, df, config["horizonte"]
    )
    
    # Generar datos para optimización
    biomasa_estimada = getBiomasa4Opti(bosque, resumen)
    
    return bosque, resumen, biomasa_final_por_rodal, biomasa_estimada




# treemun/treemun_sim/__init__.py
"""
treemun - Package for simulation of forest growth, yield and management

basic usage example:

    import treemun_sim as tm
    
    horizon=30
    stand_number=100
    
    # Forest simulation
    forest, forest_summary, last_period_biomass, collected_biomass = tm.simular_bosque(
        horizonte=horizon,
        num_rodales=stand_number
    )
    
    # Optimization
    model = tm.forest_management_optimization_model(
        forest, last_period_biomass, collected_biomass, horizon
    )
    
    results = tm.solve_model(model, 'cbc')
    solution = tm.extract_results(model, results)
    
"""

# Main functions for simulation
from .core import simular_bosque

# Optimization functions
from .optimization import (
    forest_management_optimization_model,
    solve_model,
    extract_results
)

__version__ = "1.1.0"
__author__ = "Felipe Ulloa-Fierro"

# Main functions exposed by the package
__all__ = [
    # Simulation
    "simular_bosque",
    
    # Optimization
    "forest_management_optimization_model",
    "solve_model", 
    "extract_results"
]


#!/usr/bin/env python
# coding: utf-8

# In[ ]:


# simulador_forestal/simulacion.py
"""
Módulo para la simulación del crecimiento forestal y cálculo de biomasa.
"""

import numpy as np
import pandas as pd
from typing import List, Dict, Any, Tuple

def calc_biomasa(rodal: pd.Series, e: int) -> float:
    """
    Calcula la biomasa para un rodal a una edad específica usando ecuaciones alométricas.
    """
    return max(rodal["α"] * e ** rodal["β"] + rodal["γ"], 0)

def rotular_condicion(
    edad_poda_raleo: List[int],  
    edad_cosecha: List[int],  
    edades_updated: pd.Series
) -> pd.Series:
    """
    Rotula la condición de manejo para cada edad del rodal.
    """
    condiciones = []
    
    if not edad_poda_raleo:
        edad_poda_raleo = [float('inf')]
    if not edad_cosecha:
        edad_cosecha = [float('inf')]
    
    for edad in edades_updated:
        if edad < min(edad_poda_raleo):
            condiciones.append("sin manejo")
        elif edad in edad_poda_raleo:
            condiciones.append("con manejo")
        elif edad >= max(edad_poda_raleo) and edad <= max(edad_cosecha):
            condiciones.append("con manejo")
        else:
            condiciones.append("sin manejo")

    return pd.Series(condiciones)

def generar_codigo_kitral(especie: str, edad: int, condición: str) -> str:
    """
    Genera códigos Kitral basados en especie, edad y condición de manejo.
    """
    key = (especie, edad, condición)
    
    if key[0] == "Pinus":
        if key[1] <= 3:
            value = "PL01" if key[2] == "sin manejo" else "NoCode"
        elif 3 < key[1] <= 11:
            value = "PL02" if key[2] == "sin manejo" else "PL05"
        elif 11 < key[1] <= 17:
            value = "PL03" if key[2] == "sin manejo" else "PL06"
        elif key[1] > 17:
            value = "PL04" if key[2] == "sin manejo" else "PL07"
        else:
            value = "NoCode"
    
    elif key[0] == "Eucapyltus":
        if key[1] <= 3:
            value = "PL08"
        elif 3 < key[1] <= 10:
            value = "PL09"
        else:
            value = "PL10"
    
    else:
        raise ValueError(f"Especie no soportada: {key[0]}")
    
    return value

def simula_bosque(
    rodales_con_policy: List[pd.Series],
    df: pd.DataFrame,
    horizonte: int
) -> Tuple[List[pd.DataFrame], List[Dict], Dict]:
    """
    Simula el crecimiento del bosque para todos los rodales con sus políticas aplicadas.
    
    Esta es tu función original 'simula_bosque' adaptada.
    """
    bosque = []
    resumen = []
    biomasa_final_por_rodal = {}

    for rodal in rodales_con_policy:
        tabla = pd.DataFrame()
        tabla["periodo"] = range(1, horizonte + 1)
        tabla["edadRodal"] = range(rodal["edad_in"], rodal["edad_in"] + horizonte)
        tabla["edad_updated"] = tabla["edadRodal"]

        cosecha_indices = [
            tabla[tabla["periodo"] == c].index[0] for c in rodal["cosecha"] if c in tabla["periodo"].values
        ]

        for cosecha_index in cosecha_indices:
            longitud = len(tabla.loc[cosecha_index + 1:, "edad_updated"])
            if longitud > 0:
                valores = range(1, longitud + 1)
                tabla.loc[cosecha_index + 1:, "edad_updated"] = valores

        edad_poda_raleo = [
            tabla.loc[tabla["periodo"] == pr, "edad_updated"].values[0]
            for pr in rodal["poda_raleo"] if pr in tabla["periodo"].values
        ] if isinstance(rodal["poda_raleo"], list) else []
        
        edad_cosecha = [
            tabla.loc[tabla["periodo"] == pc, "edad_updated"].values[0]
            for pc in rodal["cosecha"] if pc in tabla["periodo"].values
        ] if isinstance(rodal["cosecha"], list) else []

        tabla["condición"] = rotular_condicion(
            edad_poda_raleo, edad_cosecha, tabla["edad_updated"]
        )
        
        biomasa = []
        if rodal["TipoEspecie"] == "Eucapyltus" or pd.isnull(rodal.get("next")) or rodal["next"] == "":
            for e in tabla["edad_updated"]:
                biomasa.append(calc_biomasa(rodal, e))
        else:
            next_rodal = df[df["id"] == int(rodal["next"])].squeeze()
            poda_raleo_index = 0
            cosecha_index = 0
            usa_curva_next = rodal.get('ya_fue_raleado_al_inicio', False)

            for e in tabla["edad_updated"]:
                if cosecha_index < len(edad_cosecha) and e == edad_cosecha[cosecha_index]:
                    if usa_curva_next:
                        biomasa.append(calc_biomasa(next_rodal, e))
                    else:
                        biomasa.append(calc_biomasa(rodal, e))
                    usa_curva_next = False
                    cosecha_index += 1
                    continue
                if poda_raleo_index < len(edad_poda_raleo) and e == edad_poda_raleo[poda_raleo_index]:
                    biomasa.append(calc_biomasa(rodal, e))
                    poda_raleo_index += 1
                    usa_curva_next = True
                    continue
                if usa_curva_next:
                    biomasa.append(calc_biomasa(next_rodal, e))
                else:
                    biomasa.append(calc_biomasa(rodal, e))

        tabla["biomasa"] = (pd.Series(biomasa) * rodal["ha"]).round(3)
        tabla["id_rodal"] = rodal["id_rodal"]
        tabla["Especie"] = rodal["TipoEspecie"]

        tabla["kitral_class"] = tabla.apply(
            lambda row: generar_codigo_kitral(
                row["Especie"], row["edad_updated"], row["condición"]
            ),
            axis=1,
        )

        edad_initial = tabla.loc[0, "edad_updated"]
        edad_final = tabla.loc[horizonte - 1, "edad_updated"]

        policy_num = rodal['policy_number']
        if rodal["TipoEspecie"] == "Pinus":
            politica = f"policy_pino {policy_num}"
        elif rodal["TipoEspecie"] == "Eucapyltus":
            politica = f"policy_eucalyptus {policy_num}"
        else:
            raise ValueError(f"La especie {rodal['TipoEspecie']} no tiene políticas definidas.")
        
        tabla["politica"] = politica
        tabla["bioOPT"] = 0.0

        # Cálculo de bioOPT para podas/raleos
        for pr in rodal["poda_raleo"] if isinstance(rodal["poda_raleo"], list) else []:
            if pr in tabla["periodo"].values and pr != -1:
                valor_bio_opt = 0.0
                if pd.notnull(rodal.get("next")) and rodal["next"] != "":
                    edad_en_raleo = tabla.loc[tabla["periodo"] == pr, "edad_updated"].values[0]
                    try:
                        next_rodal_df = df[df["id"] == int(rodal["next"])]
                        if not next_rodal_df.empty:
                            next_rodal = next_rodal_df.squeeze()
                            biomasa_antes_ha = calc_biomasa(rodal, edad_en_raleo)
                            biomasa_despues_ha = calc_biomasa(next_rodal, edad_en_raleo)
                            diferencia_ha = biomasa_antes_ha - biomasa_despues_ha
                            if diferencia_ha >= 0:
                                valor_bio_opt = diferencia_ha * rodal["ha"]
                            else:
                                biomasa_total_antes = tabla.loc[tabla["periodo"] == pr, "biomasa"].values[0]
                                valor_bio_opt = biomasa_total_antes * 0.30
                        else:
                            biomasa_total_antes = tabla.loc[tabla["periodo"] == pr, "biomasa"].values[0]
                            valor_bio_opt = biomasa_total_antes * 0.30
                    except Exception:
                        biomasa_total_antes = tabla.loc[tabla["periodo"] == pr, "biomasa"].values[0]
                        valor_bio_opt = biomasa_total_antes * 0.30
                else:
                    biomasa_total_antes = tabla.loc[tabla["periodo"] == pr, "biomasa"].values[0]
                    valor_bio_opt = biomasa_total_antes * 0.30
                tabla.loc[tabla["periodo"] == pr, "bioOPT"] = valor_bio_opt
        
        # Cálculo de bioOPT para cosechas
        for pc in rodal["cosecha"] if isinstance(rodal["cosecha"], list) else []:
            if pc in tabla["periodo"].values:
                tabla.loc[tabla["periodo"] == pc, "bioOPT"] = tabla.loc[tabla["periodo"] == pc, "biomasa"].values[0]
        
        tabla = tabla.drop(columns=["edadRodal"])
        tabla = tabla.rename(columns={"edad_updated": "edad_rodal"})
        
        if horizonte in tabla["periodo"].values:
            biomasa_final_por_rodal[(rodal["id_rodal"], politica)] = (
                tabla.loc[tabla["periodo"] == horizonte, "biomasa"].values[0]
            )
        
        bosque.append(tabla)

        resumen.append(
            {
                "id_rodal": rodal["id_rodal"],
                "especie": rodal["TipoEspecie"],
                "has": rodal["ha"],
                "edad_inicial": edad_initial,
                "edad_final": edad_final,
                "policy": politica,
                "ecuacion_inicial_id": rodal["eq_id"]
            }
        )

    return bosque, resumen, biomasa_final_por_rodal

def getBiomasa4Opti(bosque: List[pd.DataFrame], resumen: List[Dict]) -> Dict:
    """
    Prepara los datos de biomasa para optimización.
    
    Esta es tu función original 'getBiomasa4Opti'.
    """
    biomasa_list = [df["bioOPT"].values for df in bosque]
    biomasa_df = pd.DataFrame(biomasa_list).T
    
    id_rodales = [item['id_rodal'] for item in resumen]
    policies = [item['policy'] for item in resumen]
    especies = [item["especie"] for item in resumen]

    df_test = pd.DataFrame([id_rodales, policies, especies])
    biomasa_df.columns = df_test.columns
    
    combined_df = pd.concat([df_test, biomasa_df], axis=0)
    combined_df.index = ['id_rodal', 'policy', "especie"] + [i+1 for i in range(len(biomasa_df))]
    
    combined_df.columns = pd.MultiIndex.from_arrays([combined_df.iloc[0], combined_df.iloc[1], combined_df.iloc[2]])
    combined_df = combined_df.drop(combined_df.index[[0, 1, 2]])

    # Conversión a diccionario para optimización
    a_I_J_dict = {
        k: v for k, v in 
        combined_df.stack()
                   .stack()
                   .stack()
                   .to_dict().items()
        if not pd.isna(v) and v > 0  # Filtrar valores nulos y ceros
    }
    
    return a_I_J_dict

def simula_bosque_completo(
    config: Dict[str, Any],
    df: pd.DataFrame,
    dict_idx: Dict
) -> Tuple[List[pd.DataFrame], List[Dict], Dict]:
    """
    Función que integra todo el proceso de simulación.
    
    Equivale a ejecutar secuencialmente:
    1. generar_rodales()
    2. generar_rodalesconpolicy() 
    3. simula_bosque()
    """
    from .generadores import generar_rodales, generar_rodalesconpolicy
    
    # Generar rodales base
    rodales = generar_rodales(config, df, dict_idx)
    
    # Aplicar políticas
    rodales_con_policy = generar_rodalesconpolicy(rodales, config)
    
    # Simular crecimiento
    bosque, resumen, biomasa_final_por_rodal = simula_bosque(
        rodales_con_policy, df, config["horizonte"]
    )
    
    return bosque, resumen, biomasa_final_por_rodal




#!/usr/bin/env python
# coding: utf-8

# In[ ]:


# simulador_forestal/generadores.py
"""
Módulo para generación de rodales aleatorios y aplicación de políticas de manejo.
"""

import numpy as np
import pandas as pd
import random
from typing import List, Tuple, Dict, Any

# IDs válidos para pinos iniciales (del código original)
IDS_PINOS_INICIALES_VALIDOS = [21, 22, 25, 26, 29, 30]

def contar_politicas_factibles(especie: str, edad_inicial: int, horizonte: int, policies_pino: List[Tuple[int, int]], policies_eucalyptus: List[Tuple[int,]]) -> int:
    """
    Cuenta cuántas políticas de manejo son factibles para un rodal
    dada su especie y edad inicial.
    """
    if especie == "Pinus":
        policies = policies_pino
    elif especie == "Eucapyltus":
        policies = policies_eucalyptus
    else:
        return 0

    contador_factibles = 0
    for policy in policies:
        es_factible = True
        if especie == "Pinus":
            if policy[0] - edad_inicial + 1 > horizonte and edad_inicial <= policy[0]:
                es_factible = False
            if edad_inicial > policy[1]:
                es_factible = False
        elif especie == "Eucapyltus":
            if policy[0] - edad_inicial + 1 > horizonte and edad_inicial <= policy[0]:
                es_factible = False
            if edad_inicial > policy[0]:
                es_factible = False
        
        if es_factible:
            contador_factibles += 1
            
    return contador_factibles

def generar_combinacion_valida(df: pd.DataFrame, especie_requerida: str = None) -> Tuple:
    """
    Genera una combinación válida de parámetros de rodal basada en la lookup table.
    """
    if especie_requerida == "Pinus":
        id_elegido = random.choice(IDS_PINOS_INICIALES_VALIDOS)
        fila_elegida = df[df['id'] == id_elegido]
        combinacion = fila_elegida.index[0]
        return combinacion

    especies = df.index.get_level_values('Especie').unique().tolist()
    zonas = df.index.get_level_values('Zona').unique().tolist()
    densidades_iniciales = df.index.get_level_values('DensidadInicial').unique().tolist()
    site_indices = df.index.get_level_values('SiteIndex').unique().tolist()
    manejos = df.index.get_level_values('Manejo').unique().tolist()
    condiciones = df.index.get_level_values('Condicion').unique().tolist()

    lookup_set = set(zip(df.index.get_level_values('Especie'),
                         df.index.get_level_values('Zona'),
                         df.index.get_level_values('SiteIndex'),
                         df.index.get_level_values('Manejo'),
                         df.index.get_level_values('Condicion'),
                         df.index.get_level_values('DensidadInicial')))

    while True:
        if especie_requerida:
            especie = especie_requerida
        else:
            especie = np.random.choice(especies)
        
        zona = np.random.choice(zonas)
        site_index = np.random.choice(site_indices)
        manejo = np.random.choice(manejos)
        condicion = np.random.choice(condiciones)
        densidad_inicial = np.random.choice(densidades_iniciales)

        combinacion = (especie, zona, site_index, manejo, condicion, densidad_inicial)

        if combinacion in lookup_set:
            if especie == "Pinus":
                try:
                    row = df.loc[combinacion]
                    if not pd.isnull(row.get("next", np.nan)):
                        return combinacion
                except:
                    continue
            else:
                return combinacion

def generar_rodales_aleatorios(
    df: pd.DataFrame,
    dict_idx: Dict,
    num_rodales: int,
    horizonte: int,
    policies_pino: List[Tuple[int, int]],
    policies_eucalyptus: List[Tuple[int,]]
) -> Dict[str, Any]:
    """
    Genera rodales aleatorios con proporción controlada de especies.
    
    Basado en la lógica original del código, genera entre 40-60% de pinos
    y el resto eucaliptos, asegurando que cada rodal tenga al menos 2 políticas factibles.
    """
    
    print(f"Generando {num_rodales} rodales con una proporción controlada (máx 60/40)...")

    # Determinar objetivos para cada especie (40-60% pinos)
    target_pinos = random.randint(int(num_rodales * 0.4), int(num_rodales * 0.6))
    target_eucas = num_rodales - target_pinos
    print(f"Objetivo de la simulación: {target_pinos} rodales de Pino, {target_eucas} rodales de Eucalipto.")

    # Listas para almacenar los datos de los rodales válidos
    edades_iniciales = []
    hectareas = []
    ids_rodales = []
    especies_aleatorias = []
    zonas_aleatorias = []
    site_indices_aleatorios = []
    densidades_iniciales_aleatorias = []
    manejos_aleatorios = []
    condiciones_aleatorias = []

    # Contadores
    pinos_generados = 0
    eucas_generados = 0

    # Bucle principal de generación
    while len(ids_rodales) < num_rodales:
        
        # Decidir qué especie generar según las cuotas
        if pinos_generados >= target_pinos:
            especie_aleatoria = "Eucapyltus"
        elif eucas_generados >= target_eucas:
            especie_aleatoria = "Pinus"
        else:
            especie_aleatoria = np.random.choice(["Pinus", "Eucapyltus"])

        # Generar edad inicial según especie
        if especie_aleatoria == "Pinus":
            edad_inicial = np.random.randint(1, 11)
        else:
            edad_inicial = np.random.randint(1, 21)

        # Verificar que tenga al menos 2 políticas factibles
        num_factibles = contar_politicas_factibles(
            especie_aleatoria, edad_inicial, horizonte, policies_pino, policies_eucalyptus
        )
        
        if num_factibles < 2:
            continue

        # Generar combinación válida
        combinacion_valida = generar_combinacion_valida(df, especie_requerida=especie_aleatoria)
        especie, zona, site_index, manejo, condicion, densidad_inicial = combinacion_valida
        hectarea = round(np.random.uniform(1, 20), 2)

        # Añadir el rodal válido a nuestras listas
        num_actual = len(ids_rodales) + 1
        ids_rodales.append(f"stand{num_actual}")
        edades_iniciales.append(edad_inicial)
        hectareas.append(hectarea)
        especies_aleatorias.append(especie)
        zonas_aleatorias.append(zona)
        site_indices_aleatorios.append(site_index)
        densidades_iniciales_aleatorias.append(densidad_inicial)
        manejos_aleatorios.append(manejo)
        condiciones_aleatorias.append(condicion)
        
        # Actualizar contadores
        if especie_aleatoria == "Pinus":
            pinos_generados += 1
        else:
            eucas_generados += 1
    
    print(f"¡Generación completa! Se crearon {pinos_generados} rodales de Pino y {eucas_generados} de Eucalipto.")

    # Crear el diccionario de configuración
    config = {
        "horizonte": horizonte,
        "rodales": num_rodales,
        "edades": edades_iniciales,
        "has": hectareas,
        "especie": especies_aleatorias,
        "siteIndex": site_indices_aleatorios,
        "zona": zonas_aleatorias,
        "densidadInicial": densidades_iniciales_aleatorias,
        "manejo": manejos_aleatorios,
        "condición": condiciones_aleatorias,
        "id_rodal": ids_rodales,
        "num_policies_pino": len(policies_pino),
        "policies_pino": policies_pino,
        "num_policies_eucalyptus": len(policies_eucalyptus),
        "policies_eucalyptus": policies_eucalyptus,
    }

    return config

def generar_rodales(config: Dict[str, Any], df: pd.DataFrame, dict_idx: Dict) -> List[pd.Series]:
    """
    Genera lista de rodales como Series de pandas basado en la configuración.
    """
    rodales = []
    for i in range(config["rodales"]):
        key = (
            config["especie"][i],
            config["zona"][i],
            config["siteIndex"][i],
            config["manejo"][i],
            config["condición"][i],
            config["densidadInicial"][i],
        )
        if key not in dict_idx:
            raise KeyError(
                f"No existe un índice para la combinación de claves {key} en la lookup table."
            )
        init_ages = config["edades"][i]
        hectareas = config["has"][i]
        rodal_id = config["id_rodal"][i]
        especie = config["especie"][i]
        rodal = pd.concat(
            (
                df.loc[key],
                pd.Series(
                    {
                        "eq_id": dict_idx[key],
                        "edad_in": init_ages,
                        "ha": hectareas,
                        "id_rodal": rodal_id,
                        "TipoEspecie": especie,
                    }
                ),
            )
        )
        rodales.append(rodal)
    return rodales

def generar_rodalesconpolicy(
    rodales: List[pd.Series], 
    config: Dict[str, Any]
) -> List[pd.Series]:
    """
    Genera rodales con políticas de manejo aplicadas.
    """
    horizonte = config["horizonte"]
    rodales_con_policy = []
    
    for rodal in rodales:
        if rodal["TipoEspecie"] == "Pinus":
            policies = config["policies_pino"]
        elif rodal["TipoEspecie"] == "Eucapyltus":
            policies = config["policies_eucalyptus"]
        else:
            continue
        
        for i, policy in enumerate(policies):
            es_factible = True
            if rodal["TipoEspecie"] == "Pinus":
                if policy[0] - rodal["edad_in"] + 1 > horizonte and rodal["edad_in"] <= policy[0]:
                    es_factible = False
                if rodal["edad_in"] > policy[1]:
                    es_factible = False
            elif rodal["TipoEspecie"] == "Eucapyltus":
                if policy[0] - rodal["edad_in"] + 1 > horizonte and rodal["edad_in"] <= policy[0]:
                    es_factible = False
                if rodal["edad_in"] > policy[0]:
                    es_factible = False
            
            if not es_factible:
                continue

            rodal_policy = rodal.copy()
            rodal_policy['policy_number'] = i + 1

            rodal_policy['ya_fue_raleado_al_inicio'] = False
            if rodal["TipoEspecie"] == "Pinus":
                if rodal["edad_in"] > policy[0]:
                    rodal_policy['ya_fue_raleado_al_inicio'] = True
            
            poda_raleo_periodos = []
            cosecha_periodos = []
            edad_rodal = rodal["edad_in"]

            for periodo in range(1, horizonte + 1):
                if rodal["TipoEspecie"] == "Pinus":
                    if edad_rodal == policy[0]:
                        poda_raleo_periodos.append(periodo)
                    if edad_rodal == policy[1]:
                        cosecha_periodos.append(periodo)
                        edad_rodal = 0
                elif rodal["TipoEspecie"] == "Eucapyltus":
                    if edad_rodal == policy[0]:
                        cosecha_periodos.append(periodo)
                        edad_rodal = 0
                edad_rodal += 1
            
            rodal_policy["poda_raleo"] = poda_raleo_periodos
            rodal_policy["cosecha"] = cosecha_periodos
            rodal_policy["horizonte"] = horizonte
            rodales_con_policy.append(rodal_policy)

    return rodales_con_policy




# treemun/setup.py

from setuptools import setup, find_packages
import os

# read the README for longer description
this_directory = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = "Growth, yield, and management simulator for Chilean plantation forests"

setup(
    name="treemun-sim",
    version="1.1.0", 
    author="Felipe Ulloa-Fierro",
    author_email="felipe.ulloa@utalca.cl", 
    description="Growth, yield, and management simulator for Chilean plantation forests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fulloaf/treemun",  
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.8",
    install_requires=[
        "numpy>=1.19.0",
        "pandas>=1.2.0",
        "pyomo>=6.0.0",  # new dependency for optimization
    ],
    include_package_data=True,
    package_data={
        "treemun_sim": ["data/*.csv"],
    },
    # optional dependencies for developtment and solvers 
    extras_require={
        "dev": [
            "pytest>=6.0",
            "pytest-cov>=2.0",
            "black>=21.0",
            "flake8>=3.8",
        ],
        "solvers": [
            "pulp>=2.0",  # include CBC (free solver)
        ],
        "solvers-extended": [
            "pulp>=2.0",  # CBC solver
            "cplex",      # CPLEX (requires license) 
        ],
        "complete": [
            "pulp>=2.0",
            "pytest>=6.0",
            "pytest-cov>=2.0",
            "black>=21.0",
            "flake8>=3.8",
        ],
    },
    # Uptdated metadata
    keywords="plantation forest, simulation, biomass growth, management policies, optimization, forest management",
    project_urls={
        "Bug Reports": "https://github.com/fulloaf/treemun/issues",
        "Source": "https://github.com/fulloaf/treemun",
        "Documentation": "https://github.com/fulloaf/treemun#readme",
    },
)


# treemun/treemun_sim/optimization.py
"""
Optimization module for treemun forest management models.
Provides functions to create, solve, and extract results from a forest optimization model example.
"""

from pyomo.environ import *
from pyomo.opt import SolverFactory
import os

def forest_management_optimization_model(bosque, a_i_j_T, a_i_j_t, horizon, 
                                       pine_revenue=9, eucalyptus_revenue=10, 
                                       min_ending_biomass=30000, discount_rate=0.08):
    """
    Creates and returns a forest management optimization model using Pyomo.
    
    Args:
        bosque (List[pd.DataFrame]): List of forest simulation DataFrames, each representing 
                                   a stand-policy combination with temporal dynamics
        a_i_j_T (Dict): Dictionary with final standing biomass (at the end of the planning horizon)
                                values by stand-policy combination, structure: {(stand_id, policy): biomass_value}
        a_i_j_t (Dict): Dictionary with harvestable biomass coefficients by period of thinning/harvest respectively,
                       structure: {(period, species, policy, stand_id): biomass_value}
        horizon (int): Planning horizon in years
        pine_revenue (float or List[float]): biomass revenue for pine species ($/m³). 
                                         Can be a single value (constant revenue) or 
                                         a list of length horizon (variable revenue by period)
        eucalyptus_revenue (float or List[float]): biomass revenue for eucalyptus species ($/m³). 
                                               Can be a single value or list of length horizon
        min_ending_biomass (float): Minimum total standing biomass required at end of horizon (m³)
        discount_rate (float): Annual discount rate for NPV calculation (decimal, e.g., 0.08 = 8%)
    
    Returns:
        model (ConcreteModel): Configured Pyomo optimization model ready for solving
    """
    
    # Validate and convert revenue to arrays
    if isinstance(pine_revenue, (int, float)):
        pine_revenue_array = [pine_revenue] * horizon
    else:
        pine_revenue_array = list(pine_revenue)
        if len(pine_revenue_array) != horizon:
            raise ValueError(f"pine_revenue must have length {horizon}, got {len(pine_revenue_array)}")
    
    if isinstance(eucalyptus_revenue, (int, float)):
        eucalyptus_revenue_array = [eucalyptus_revenue] * horizon
    else:
        eucalyptus_revenue_array = list(eucalyptus_revenue)
        if len(eucalyptus_revenue_array) != horizon:
            raise ValueError(f"eucalyptus_revenue must have length {horizon}, got {len(eucalyptus_revenue_array)}")
    
    # Extract sets from forest simulation data
    I_pino = set()
    I_euca = set()
    J_pino = set()
    J_euca = set()

    for df in bosque:
        # Extract stand IDs by species
        pinus_df = df[df["Especie"] == "Pinus"]
        I_pino.update(pinus_df['id_rodal'].unique())
        
        euca_df = df[df["Especie"] == "Eucapyltus"]
        I_euca.update(euca_df['id_rodal'].unique())
        
        # Extract policies by species
        for policy in df['politica'].unique():
            if 'pino' in policy:
                J_pino.add(policy)
            elif 'eucalyptus' in policy:
                J_euca.add(policy)

    # Create Pyomo model
    model = ConcreteModel()

    # Define sets
    model.I_pino = Set(initialize=sorted(I_pino))
    model.I_euca = Set(initialize=sorted(I_euca))
    model.J_pino = Set(initialize=sorted(J_pino))
    model.J_euca = Set(initialize=sorted(J_euca))
    model.T = RangeSet(1, horizon)
    model.Epino = Set(initialize=["Pinus"])
    model.Eeuca = Set(initialize=["Eucapyltus"])

    # Economic parameters (indexed by period for variable revenue)
    model.pine_revenue = Param(model.T, initialize={t: pine_revenue_array[t-1] for t in model.T})
    model.eucalyptus_revenue = Param(model.T, initialize={t: eucalyptus_revenue_array[t-1] for t in model.T})
    model.min_ending_biomass = Param(initialize=min_ending_biomass)
    model.discount_rate = Param(initialize=discount_rate)

    # Define valid stand-policy combinations
    valid_pino_indices = [(i, j) for i in model.I_pino for j in model.J_pino 
                          if any((t, e, j, i) in a_i_j_t for t in model.T for e in model.Epino)]

    valid_euca_indices = [(i, j) for i in model.I_euca for j in model.J_euca
                          if any((t, e, j, i) in a_i_j_t for t in model.T for e in model.Eeuca)]

    # Decision variables
    model.x_pine = Var(valid_pino_indices, within=Binary)
    model.x_eucalyptus = Var(valid_euca_indices, within=Binary)
    model.harvest_volume = Var(model.T, within=NonNegativeReals)

    # Objective function: Maximize Net Present Value
    def objective_rule(model):
        pine_npv = sum(
            a_i_j_t[(t, e, j, i)] * model.x_pine[i, j] * model.pine_revenue[t] / (1 + model.discount_rate)**t
            for t in model.T for e in model.Epino for j in model.J_pino for i in model.I_pino
            if (i, j) in valid_pino_indices and (t, e, j, i) in a_i_j_t
        )
        
        eucalyptus_npv = sum(
            a_i_j_t[(t, e, j, i)] * model.x_eucalyptus[i, j] * model.eucalyptus_revenue[t] / (1 + model.discount_rate)**t
            for t in model.T for e in model.Eeuca for j in model.J_euca for i in model.I_euca
            if (i, j) in valid_euca_indices and (t, e, j, i) in a_i_j_t
        )
        
        return pine_npv + eucalyptus_npv

    model.objective = Objective(rule=objective_rule, sense=maximize)

    # Constraint: Single policy assignment per stand
    def single_assignment_pine_rule(model, i):
        return sum(model.x_pine[i, j] for j in model.J_pino if (i, j) in valid_pino_indices) == 1

    def single_assignment_eucalyptus_rule(model, i):
        return sum(model.x_eucalyptus[i, j] for j in model.J_euca if (i, j) in valid_euca_indices) == 1

    model.single_assignment_pine = Constraint(model.I_pino, rule=single_assignment_pine_rule)
    model.single_assignment_eucalyptus = Constraint(model.I_euca, rule=single_assignment_eucalyptus_rule)

    # Constraint: Track harvest volumes by period
    def harvest_volume_tracking_rule(model, t):
        return sum(
            a_i_j_t[(t, e, j, i)] * model.x_pine[i, j]
            for e in model.Epino for j in model.J_pino for i in model.I_pino
            if (t, e, j, i) in a_i_j_t
        ) + sum(
            a_i_j_t[(t, e, j, i)] * model.x_eucalyptus[i, j]
            for e in model.Eeuca for j in model.J_euca for i in model.I_euca
            if (t, e, j, i) in a_i_j_t
        ) == model.harvest_volume[t]

    model.harvest_volume_tracking = Constraint(model.T, rule=harvest_volume_tracking_rule)

    # Constraint: Even-flow (non-decreasing harvest volumes)
    def even_flow_rule(model, t):
        if t < model.T.last():
            return model.harvest_volume[t + 1] >= model.harvest_volume[t]
        else:
            return Constraint.Skip

    model.even_flow = Constraint(model.T, rule=even_flow_rule)

    # Constraint: Minimum ending forest inventory
    def sustainability_rule(model):
        total_ending_biomass = sum(
            a_i_j_T.get((i, j), 0) * model.x_pine[i, j] 
            for (i, j) in valid_pino_indices
            if (i, j) in a_i_j_T
        ) + sum(
            a_i_j_T.get((i, j), 0) * model.x_eucalyptus[i, j]
            for (i, j) in valid_euca_indices
            if (i, j) in a_i_j_T
        )
        return total_ending_biomass >= model.min_ending_biomass

    model.sustainability = Constraint(rule=sustainability_rule)

    return model

def solve_model(model, solver_name, gap=0.01, executable_path=None, tee=True):
    """
    Solves a Pyomo model, requiring the solver to be in the system's PATH
    or for its path to be specified explicitly.
    
    Args:
        model (Pyomo Model): The optimization model to be solved.
        solver_name (str): Name of the solver ('cplex' or 'cbc').
        gap (float, optional): The relative optimality gap. Defaults to 0.01.
        executable_path (str, optional): Direct path to the solver's executable.
                                          If None, the solver is assumed to be in the system's PATH.
                                          Defaults to None.
        tee (bool, optional): If True, displays the solver's output in the console. Defaults to True.
        
    Returns:
        results: The Pyomo results object from the solver.
    """
    solver_name = solver_name.lower()
    
    # If a path is provided, verify that it exists.
    if executable_path and not os.path.exists(executable_path):
        raise FileNotFoundError(f"The specified executable was not found at: {executable_path}")
    
    # The final path is the one provided by the user, or None.
    # If it's None, Pyomo will search for the solver in the system's PATH.
    solver = SolverFactory(solver_name, executable=executable_path)
    
    # Configure solver-specific options
    if solver_name == 'cplex':
        solver.options['mipgap'] = gap
    elif solver_name == 'cbc':
        solver.options['ratioGap'] = gap  # Note: The option is named 'ratioGap' in CBC
    else:
        raise ValueError(f"Solver '{solver_name}' is not supported or the name is incorrect. Choose from 'cplex' or 'cbc'")
    
    # Solve the Model
    log_filename = f"log_{solver_name}.txt"
    print(f"--- Solving with {solver_name.upper()} | Gap: {gap*100}% ---")
    results = solver.solve(model, tee=tee, logfile=log_filename)
    print(f"--- Solving finished. Log saved to '{log_filename}' ---")
    
    return results

def extract_results(model, results):
    """
    Extracts key results from a solved Pyomo model.
    
    Args:
        model (Pyomo Model): The optimization model after being solved.
        results (Pyomo Results object): The results object from the solver.
        
    Returns:
        dict: A dictionary containing the objective value, the harvest vector, lists of 
              pine and eucalyptus stand assignments, and their respective counts. 
              Returns None if no solution was found.
    """
    # Check if the solver found an optimal or feasible solution
    term_cond = results.solver.termination_condition
    if term_cond == TerminationCondition.optimal or term_cond == TerminationCondition.feasible:
        
        # Extract the Objective Function Value
        objective_value = value(model.objective)
        
        # Build the harvest vector
        harvest_vector = [model.harvest_volume[t].value for t in model.T]
        
        # Get coordinates for x_pine where the value is greater than zero
        # A small tolerance is used to avoid floating-point precision issues.
        xpino_values = [(i, j) for (i, j) in model.x_pine if model.x_pine[i, j].value > 1e-6]
        
        # Get coordinates for x_eucalyptus where the value is greater than zero
        xeuca_values = [(i, j) for (i, j) in model.x_eucalyptus if model.x_eucalyptus[i, j].value > 1e-6]
        
        # Package everything into a dictionary for a clean and organized return
        output_data = {
            'objective_value': objective_value,
            'total_harvest_per_period': harvest_vector,
            'pinus_stand_plan': xpino_values,
            'total_pinus_stand_treated': len(xpino_values),
            'eucalyptus_stand_plan': xeuca_values,
            'total_eucalyptus_stand_treated': len(xeuca_values)
        }
        
        return output_data
        
    else:
        print(f"\nAn optimal solution was not found. Termination condition: {term_cond}")
        return None




